@@ -4,13 +4,13 @@
 //! The state of a process.
 
 use std::borrow::{Borrow, Cow};
-use std::collections::{HashMap, HashSet};
+use std::collections::{BTreeMap, BTreeSet, HashMap, HashSet};
 use std::io;
 use std::io::prelude::*;
+use std::time::SystemTime;
 
 use crate::system_info::SystemInfo;
 use crate::{FrameSymbolizer, SymbolStats};
-use chrono::prelude::*;
 use minidump::system_info::Cpu;
 use minidump::*;
 use serde_json::json;
@@ -36,35 +36,104 @@ pub enum FrameTrust {
     Context,
 }
 
+#[derive(Debug)]
+pub enum CallingConvention {
+    Cdecl,
+    WindowsThisCall,
+    OtherThisCall,
+}
+
+/// Arguments for this function
+#[derive(Debug)]
+pub struct FunctionArgs {
+    /// What we assumed the calling convention was
+    pub calling_convention: CallingConvention,
+
+    /// The actual arguments
+    pub args: Vec<FunctionArg>,
+}
+
+/// A function argument
+#[derive(Debug)]
+pub struct FunctionArg {
+    /// The name of the argument (usually actually just the type)
+    pub name: String,
+    /// The value of the argument
+    pub value: Option<u64>,
+}
+
 /// A single stack frame produced from unwinding a thread's stack.
 #[derive(Debug)]
 pub struct StackFrame {
-    // The program counter location as an absolute virtual address.
-    //
-    // - For the innermost called frame in a stack, this will be an exact
-    //   program counter or instruction pointer value.
-    //
-    // - For all other frames, this address is within the instruction that
-    //   caused execution to branch to this frame's callee (although it may
-    //   not point to the exact beginning of that instruction). This ensures
-    //   that, when we look up the source code location for this frame, we
-    //   get the source location of the call, not of the point at which
-    //   control will resume when the call returns, which may be on the next
-    //   line. (If the compiler knows the callee never returns, it may even
-    //   place the call instruction at the very end of the caller's machine
-    //   code, such that the "return address" (which will never be used)
-    //   immediately after the call instruction is in an entirely different
-    //   function, perhaps even from a different source file.)
-    //
-    // On some architectures, the return address as saved on the stack or in
-    // a register is fine for looking up the point of the call. On others, it
-    // requires adjustment. ReturnAddress returns the address as saved by the
-    // machine.
+    /// The program counter location as an absolute virtual address.
+    ///
+    /// - For the innermost called frame in a stack, this will be an exact
+    ///   program counter or instruction pointer value.
+    ///
+    /// - For all other frames, this address is within the instruction that
+    ///   caused execution to branch to this frame's callee (although it may
+    ///   not point to the exact beginning of that instruction). This ensures
+    ///   that, when we look up the source code location for this frame, we
+    ///   get the source location of the call, not of the point at which
+    ///   control will resume when the call returns, which may be on the next
+    ///   line. (If the compiler knows the callee never returns, it may even
+    ///   place the call instruction at the very end of the caller's machine
+    ///   code, such that the "return address" (which will never be used)
+    ///   immediately after the call instruction is in an entirely different
+    ///   function, perhaps even from a different source file.)
+    ///
+    /// On some architectures, the return address as saved on the stack or in
+    /// a register is fine for looking up the point of the call. On others, it
+    /// requires adjustment.
     pub instruction: u64,
 
-    // The module in which the instruction resides.
+    /// The instruction address (program counter) that execution of this function
+    /// would resume at, if the callee returns.
+    ///
+    /// This is exactly **the return address of the of the callee**. We use this
+    /// nonstandard terminology because just calling this "return address"
+    /// would be ambiguous and too easy to mix up.
+    ///
+    /// **Note:** you should strongly prefer using [`StackFrame::instruction`][], which should
+    /// be the address of the instruction before this one which called the callee.
+    /// That is the instruction that this function was logically "executing" when the
+    /// program's state was captured, and therefore what people expect from
+    /// backtraces.
+    ///
+    /// This is more than a matter of user expections: **there are situations
+    /// where this value is nonsensical but the [`StackFrame::instruction`][] is valid.**
+    ///
+    /// Specifically, if the callee is "noreturn" then *this function should
+    /// never resume execution*. The compiler has no obligation to emit any
+    /// instructions after such a CALL, but CALL still implicitly pushes the
+    /// instruction after itself to the stack. Such a return address may
+    /// therefore be outside the "bounds" of this function!!!
+    ///
+    /// Yes, compilers *can* just immediately jump into the callee for
+    /// noreturn calls, but it's genuinely very helpful for them to emit a
+    /// CALL because it keeps the stack reasonable for backtraces and
+    /// debuggers, which are more interested in [`StackFrame::instruction`][] anyway!
+    ///
+    /// (If this is the top frame of the call stack, then `resume_address`
+    /// and `instruction` are exactly equal and should reflect the actual
+    /// program counter of this thread.)
+    pub resume_address: u64,
+
+    /// The module in which the instruction resides.
     pub module: Option<MinidumpModule>,
 
+    /// Any unloaded modules which overlap with this address.
+    ///
+    /// This is currently only populated if `module` is None.
+    ///
+    /// Since unloaded modules may overlap, there may be more than
+    /// one module. Since a module may be unloaded and reloaded at
+    /// multiple positions, we keep track of all the offsets that
+    /// apply. BTrees are used to produce a more stable output.
+    ///
+    /// So this is a `BTreeMap<module_name, Set<offsets>>`.
+    pub unloaded_modules: BTreeMap<String, BTreeSet<u64>>,
+
     /// The function name, may be omitted if debug symbols are not available.
     pub function_name: Option<String>,
 
@@ -93,6 +162,9 @@ pub struct StackFrame {
 
     /// The CPU context containing register state for this frame.
     pub context: MinidumpContext,
+
+    /// Any function args we recovered.
+    pub arguments: Option<FunctionArgs>,
 }
 
 /// Information about the results of unwinding a thread's stack.
@@ -119,6 +191,8 @@ pub struct CallStack {
     pub frames: Vec<StackFrame>,
     /// Information about this `CallStack`.
     pub info: CallStackInfo,
+    /// The identifier of the thread.
+    pub thread_id: u32,
     /// The name of the thread, if known.
     pub thread_name: Option<String>,
     /// The GetLastError() value stored in the TEB.
@@ -138,9 +212,9 @@ pub struct ProcessState {
     /// The PID of the process.
     pub process_id: Option<u32>,
     /// When the minidump was written.
-    pub time: DateTime<Utc>,
+    pub time: SystemTime,
     /// When the process started, if available
-    pub process_create_time: Option<DateTime<Utc>>,
+    pub process_create_time: Option<SystemTime>,
     /// Known code signing certificates (module name => cert name)
     pub cert_info: HashMap<String, String>,
     /// If the process crashed, a `CrashReason` describing the crash reason.
@@ -219,23 +293,21 @@ impl StackFrame {
     pub fn from_context(context: MinidumpContext, trust: FrameTrust) -> StackFrame {
         StackFrame {
             instruction: context.get_instruction_pointer(),
+            // Initialized the same as `instruction`, but left unmodified during stack walking.
+            resume_address: context.get_instruction_pointer(),
             module: None,
+            unloaded_modules: BTreeMap::new(),
             function_name: None,
             function_base: None,
             parameter_size: None,
             source_file_name: None,
             source_line: None,
             source_line_base: None,
+            arguments: None,
             trust,
             context,
         }
     }
-
-    /// Return the actual return address, as saved on the stack or in a
-    /// register. See the comments for `StackFrame::instruction` for details.
-    pub fn return_address(&self) -> u64 {
-        self.instruction
-    }
 }
 
 impl FrameSymbolizer for StackFrame {
@@ -313,10 +385,11 @@ fn json_registers(ctx: &MinidumpContext) -> serde_json::Value {
 
 impl CallStack {
     /// Create a `CallStack` with `info` and no frames.
-    pub fn with_info(info: CallStackInfo) -> CallStack {
+    pub fn with_info(id: u32, info: CallStackInfo) -> CallStack {
         CallStack {
             info,
             frames: vec![],
+            thread_id: id,
             thread_name: None,
             last_error_value: None,
         }
@@ -362,11 +435,63 @@ impl CallStack {
                     write!(f, " + {:#x}", addr - module.base_address())?;
                 }
             } else {
-                writeln!(f, "{:#x}", addr)?;
+                write!(f, "{:#x}", addr)?;
+
+                // List off overlapping unloaded modules.
+
+                // First we need to collect them up by name so that we can print
+                // all the overlaps from one module together and dedupe them.
+
+                for (name, offsets) in &frame.unloaded_modules {
+                    write!(f, " (unloaded {}@", name)?;
+                    let mut first = true;
+                    for offset in offsets {
+                        if first {
+                            write!(f, "{:#x}", offset)?;
+                        } else {
+                            // `|` is our separator for multiple entries
+                            write!(f, "|{:#x}", offset)?;
+                        }
+                        first = false;
+                    }
+                    write!(f, ")")?;
+                }
             }
             writeln!(f)?;
             print_registers(f, &frame.context)?;
             writeln!(f, "    Found by: {}", frame.trust.description())?;
+
+            if let Some(args) = &frame.arguments {
+                use MinidumpRawContext::*;
+                let pointer_width = match &frame.context.raw {
+                    X86(_) | Ppc(_) | Sparc(_) | Arm(_) | Mips(_) => 4,
+                    Ppc64(_) | Amd64(_) | Arm64(_) | OldArm64(_) => 8,
+                };
+
+                let cc_summary = match args.calling_convention {
+                    CallingConvention::Cdecl => "cdecl [static function]",
+                    CallingConvention::WindowsThisCall => "windows thiscall [C++ member function]",
+                    CallingConvention::OtherThisCall => {
+                        "non-windows thiscall [C++ member function]"
+                    }
+                };
+
+                writeln!(f, "    Arguments (assuming {})", cc_summary)?;
+                for (idx, arg) in args.args.iter().enumerate() {
+                    if let Some(val) = arg.value {
+                        if pointer_width == 4 {
+                            writeln!(f, "        arg {} ({}) = 0x{:08x}", idx, arg.name, val)?;
+                        } else {
+                            writeln!(f, "        arg {} ({}) = 0x{:016x}", idx, arg.name, val)?;
+                        }
+                    } else {
+                        writeln!(f, "        arg {} ({}) = <unknown>", idx, arg.name)?;
+                    }
+                }
+                // Add an extra new-line between frames when there's function arguments to make
+                // it more readable.
+                writeln!(f)?;
+            }
         }
         Ok(())
     }
@@ -401,7 +526,7 @@ impl ProcessState {
 
     fn print_internal<T: Write>(&self, f: &mut T, brief: bool) -> io::Result<()> {
         writeln!(f, "Operating system: {}", self.system_info.os.long_name())?;
-        if let Some(ref ver) = self.system_info.os_version {
+        if let Some(ref ver) = self.system_info.format_os_version() {
             writeln!(f, "                  {}", ver)?;
         }
         writeln!(f, "CPU: {}", self.system_info.cpu)?;
@@ -474,8 +599,8 @@ Crash address: {:#x}
             writeln!(f)?;
         }
         if let Some(ref time) = self.process_create_time {
-            let uptime = self.time - *time;
-            writeln!(f, "Process uptime: {} seconds", uptime.num_seconds())?;
+            let uptime = self.time.duration_since(*time).unwrap_or_default();
+            writeln!(f, "Process uptime: {} seconds", uptime.as_secs())?;
         } else {
             writeln!(f, "Process uptime: not available")?;
         }
@@ -605,7 +730,7 @@ Unknown streams encountered:
 
     /// Outputs json in a schema compatible with mozilla's Socorro crash reporting servers.
     ///
-    #[doc = include_str!("../json-schema.md")]
+    /// See the top level documentation of this library for the stable JSON schema.
     pub fn print_json<T: Write>(&self, f: &mut T, pretty: bool) -> Result<(), serde_json::Error> {
         // See ../json-schema.md for details on this format.
 
@@ -621,7 +746,7 @@ Unknown streams encountered:
             "system_info": {
                 // Linux | Windows NT | Mac OS X
                 "os": sys.os.long_name(),
-                "os_ver": sys.os_version,
+                "os_ver": sys.format_os_version(),
                 // x86 | amd64 | arm | ppc | sparc
                 "cpu_arch": sys.cpu.to_string(),
                 "cpu_info": sys.cpu_info,
@@ -662,6 +787,7 @@ Unknown streams encountered:
 
             // the first module is always the main one
             "main_module": 0,
+            // [UNSTABLE:evil_json]
             "modules_contains_cert_info": !self.cert_info.is_empty(),
             "modules": self.modules.iter().map(|module| {
                 let full_name = module.code_file();
@@ -681,11 +807,12 @@ Unknown streams encountered:
                     // filename | empty string
                     "debug_file": basename(module.debug_file().unwrap_or(Cow::Borrowed("")).borrow()),
                     // [[:xdigit:]]{33} | empty string
-                    "debug_id": module.debug_identifier().unwrap_or(Cow::Borrowed("")),
+                    "debug_id": module.debug_identifier().unwrap_or_default().breakpad().to_string(),
                     "end_addr": json_hex(module.raw.base_of_image + module.raw.size_of_image as u64),
                     "filename": &name,
-                    "code_id": module.code_identifier(),
+                    "code_id": module.code_identifier().unwrap_or_default().as_str(),
                     "version": module.version(),
+                    // [UNSTABLE:evil_json]
                     "cert_subject": self.cert_info.get(name),
 
                     // These are all just metrics for debugging minidump-processor's execution
@@ -726,6 +853,15 @@ Unknown streams encountered:
                         .map(|module| frame.instruction - module.raw.base_of_image)
                         .map(json_hex),
                     // optional
+                    "unloaded_modules": if frame.unloaded_modules.is_empty() {
+                        None
+                    } else {
+                        Some(frame.unloaded_modules.iter().map(|(module, offsets)| json!({
+                            "module": module,
+                            "offsets": offsets.iter().copied().map(json_hex).collect::<Vec<_>>(),
+                        })).collect::<Vec<_>>())
+                    },
+                    // optional
                     "function_offset": frame
                         .function_base
                         .map(|func_base| frame.instruction - func_base)
@@ -738,17 +874,11 @@ Unknown streams encountered:
 
             "unloaded_modules": self.unloaded_modules.iter().map(|module| json!({
                 "base_addr": json_hex(module.raw.base_of_image),
-                "code_id": module.code_identifier(),
+                "code_id": module.code_identifier().unwrap_or_default().as_str(),
                 "end_addr": json_hex(module.raw.base_of_image + module.raw.size_of_image as u64),
                 "filename": module.name,
                 "cert_subject": self.cert_info.get(&module.name),
             })).collect::<Vec<_>>(),
-
-            "sensitive": {
-                // TODO: Issue #25
-                // low | medium | high | interesting | none | ERROR: *
-                "exploitability": null,
-            }
         });
 
         if let Some(requesting_thread) = self.requesting_thread {
@@ -760,26 +890,30 @@ Unknown streams encountered:
             // of a normal "threads" entry, while the original schema strips
             // many of the fields here. We don't to keep things more uniform.
 
-            let registers = json_registers(&self.threads[requesting_thread].frames[0].context);
+            // We can't do any of this work if we don't have at least one frame.
+            if let Some(f) = self.threads[requesting_thread].frames.get(0) {
+                let registers = json_registers(&f.context);
 
-            // Yuck, spidering through json...
-            let mut thread =
-                output.get_mut("threads").unwrap().as_array().unwrap()[requesting_thread].clone();
-            let thread_obj = thread.as_object_mut().unwrap();
-            let frames = thread_obj
-                .get_mut("frames")
-                .unwrap()
-                .as_array_mut()
-                .unwrap();
-            let frame = frames[0].as_object_mut().unwrap();
+                // Yuck, spidering through json...
+                let mut thread = output.get_mut("threads").unwrap().as_array().unwrap()
+                    [requesting_thread]
+                    .clone();
+                let thread_obj = thread.as_object_mut().unwrap();
+                let frames = thread_obj
+                    .get_mut("frames")
+                    .unwrap()
+                    .as_array_mut()
+                    .unwrap();
+                let frame = frames[0].as_object_mut().unwrap();
 
-            frame.insert(String::from("registers"), registers);
-            thread_obj.insert(String::from("threads_index"), json!(requesting_thread));
+                frame.insert(String::from("registers"), registers);
+                thread_obj.insert(String::from("threads_index"), json!(requesting_thread));
 
-            output
-                .as_object_mut()
-                .unwrap()
-                .insert(String::from("crashing_thread"), thread);
+                output
+                    .as_object_mut()
+                    .unwrap()
+                    .insert(String::from("crashing_thread"), thread);
+            }
         }
 
         if pretty {
