@@ -3,8 +3,9 @@
 
 use crate::process_state::*;
 use crate::stackwalker::walk_stack;
-use crate::{string_symbol_supplier, Symbolizer};
+use crate::{string_symbol_supplier, Symbolizer, SystemInfo};
 use minidump::format::CONTEXT_X86;
+use minidump::system_info::{Cpu, Os};
 use minidump::*;
 use std::collections::HashMap;
 use test_assembler::*;
@@ -29,7 +30,7 @@ impl TestFixture {
         }
     }
 
-    pub fn walk_stack(&self, stack: Section) -> CallStack {
+    pub async fn walk_stack(&self, stack: Section) -> CallStack {
         let context = MinidumpContext {
             raw: MinidumpRawContext::X86(self.raw.clone()),
             valid: MinidumpContextValidity::All,
@@ -43,13 +44,24 @@ impl TestFixture {
             size,
             bytes: &stack,
         };
+        let system_info = SystemInfo {
+            os: Os::Windows,
+            os_version: None,
+            os_build: None,
+            cpu: Cpu::X86,
+            cpu_info: None,
+            cpu_microcode_version: None,
+            cpu_count: 1,
+        };
         let symbolizer = Symbolizer::new(string_symbol_supplier(self.symbols.clone()));
         walk_stack(
             &Some(&context),
             Some(&stack_memory),
             &self.modules,
+            &system_info,
             &symbolizer,
         )
+        .await
     }
 
     pub fn add_symbols(&mut self, name: String, symbols: String) {
@@ -57,15 +69,15 @@ impl TestFixture {
     }
 }
 
-#[test]
-fn test_simple() {
+#[tokio::test]
+async fn test_simple() {
     let mut f = TestFixture::new();
     let mut stack = Section::new();
     stack.start().set_const(0x80000000);
     stack = stack.D32(0).D32(0); // end-of-stack marker
     f.raw.eip = 0x40000200;
     f.raw.ebp = 0x80000000;
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 1);
     let f = &s.frames[0];
     let m = f.module.as_ref().unwrap();
@@ -75,8 +87,8 @@ fn test_simple() {
 // Walk a traditional frame. A traditional frame saves the caller's
 // %ebp just below the return address, and has its own %ebp pointing
 // at the saved %ebp.
-#[test]
-fn test_traditional() {
+#[tokio::test]
+async fn test_traditional() {
     let mut f = TestFixture::new();
     let frame0_ebp = Label::new();
     let frame1_ebp = Label::new();
@@ -86,7 +98,7 @@ fn test_traditional() {
         .append_repeated(12, 0) // frame 0: space
         .mark(&frame0_ebp) // frame 0 %ebp points here
         .D32(&frame1_ebp) // frame 0: saved %ebp
-        .D32(0x40008679) // frame 0: return address
+        .D32(0x40008679) // frame 0: resume address
         .append_repeated(8, 0) // frame 1: space
         .mark(&frame1_ebp) // frame 1 %ebp points here
         .D32(0) // frame 1: saved %ebp (stack end)
@@ -94,13 +106,14 @@ fn test_traditional() {
     f.raw.eip = 0x4000c7a5;
     f.raw.esp = stack.start().value().unwrap() as u32;
     f.raw.ebp = frame0_ebp.value().unwrap() as u32;
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 2);
     {
         let f0 = &s.frames[0];
         assert_eq!(f0.trust, FrameTrust::Context);
         assert_eq!(f0.context.valid, MinidumpContextValidity::All);
         assert_eq!(f0.instruction, 0x4000c7a5);
+        assert_eq!(f0.resume_address, 0x4000c7a5);
         // eip
         // ebp
     }
@@ -109,6 +122,7 @@ fn test_traditional() {
         assert_eq!(f1.trust, FrameTrust::FramePointer);
         // ContextValidity
         assert_eq!(f1.instruction, 0x40008678);
+        assert_eq!(f1.resume_address, 0x40008679);
         // eip
         // ebp
     }
@@ -116,8 +130,8 @@ fn test_traditional() {
 
 // Walk a traditional frame, but use a bogus %ebp value, forcing a scan
 // of the stack for something that looks like a return address.
-#[test]
-fn test_traditional_scan() {
+#[tokio::test]
+async fn test_traditional_scan() {
     let mut f = TestFixture::new();
     let frame1_esp = Label::new();
     let frame1_ebp = Label::new();
@@ -144,7 +158,7 @@ fn test_traditional_scan() {
     // for something that looks like a return address.
     f.raw.ebp = 0xd43eed6e;
 
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 2);
 
     {
@@ -153,6 +167,7 @@ fn test_traditional_scan() {
         assert_eq!(f0.trust, FrameTrust::Context);
         assert_eq!(f0.context.valid, MinidumpContextValidity::All);
         assert_eq!(f0.instruction, 0x4000f49d);
+        assert_eq!(f0.resume_address, 0x4000f49d);
 
         if let MinidumpRawContext::X86(ctx) = &f0.context.raw {
             assert_eq!(ctx.eip, 0x4000f49d);
@@ -175,6 +190,7 @@ fn test_traditional_scan() {
             unreachable!();
         }
         assert_eq!(f1.instruction + 1, 0x4000129d);
+        assert_eq!(f1.resume_address, 0x4000129d);
 
         if let MinidumpRawContext::X86(ctx) = &f1.context.raw {
             assert_eq!(ctx.eip, 0x4000129d);
@@ -187,8 +203,8 @@ fn test_traditional_scan() {
 }
 
 // Force scanning for a return address a long way down the stack
-#[test]
-fn test_traditional_scan_long_way() {
+#[tokio::test]
+async fn test_traditional_scan_long_way() {
     let mut f = TestFixture::new();
     let frame1_esp = Label::new();
     let frame1_ebp = Label::new();
@@ -217,7 +233,7 @@ fn test_traditional_scan_long_way() {
     // for something that looks like a return address.
     f.raw.ebp = 0xd43eed6e;
 
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 2);
 
     {
@@ -306,13 +322,13 @@ fn init_cfi_state() -> (TestFixture, Section, CONTEXT_X86, MinidumpContextValidi
     (f, stack, expected, expected_valid)
 }
 
-fn check_cfi(
+async fn check_cfi(
     f: TestFixture,
     stack: Section,
     expected: CONTEXT_X86,
     expected_valid: MinidumpContextValidity,
 ) {
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 2);
 
     {
@@ -350,8 +366,8 @@ fn check_cfi(
     unreachable!();
 }
 
-#[test]
-fn test_cfi_at_4000() {
+#[tokio::test]
+async fn test_cfi_at_4000() {
     let (mut f, mut stack, mut expected, expected_valid) = init_cfi_state();
 
     let frame1_rsp = Label::new();
@@ -363,11 +379,11 @@ fn test_cfi_at_4000() {
     expected.set_register("esp", frame1_rsp.value().unwrap() as u32);
     f.raw.set_register("eip", 0x40004000);
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_at_4001() {
+#[tokio::test]
+async fn test_cfi_at_4001() {
     let (mut f, mut stack, mut expected, expected_valid) = init_cfi_state();
 
     let frame1_rsp = Label::new();
@@ -381,11 +397,11 @@ fn test_cfi_at_4001() {
     f.raw.set_register("eip", 0x40004001);
     f.raw.set_register("ebx", 0x91aa9a8b);
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_at_4002() {
+#[tokio::test]
+async fn test_cfi_at_4002() {
     let (mut f, mut stack, mut expected, expected_valid) = init_cfi_state();
 
     let frame1_rsp = Label::new();
@@ -400,11 +416,11 @@ fn test_cfi_at_4002() {
     f.raw.set_register("ebx", 0x53d1379d);
     f.raw.set_register("esi", 0xa5c790ed);
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_at_4003() {
+#[tokio::test]
+async fn test_cfi_at_4003() {
     let (mut f, mut stack, mut expected, expected_valid) = init_cfi_state();
 
     let frame1_rsp = Label::new();
@@ -423,11 +439,11 @@ fn test_cfi_at_4003() {
     f.raw.set_register("esi", 0xa97f229d);
     f.raw.set_register("edi", 0xb05cc997);
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_at_4004() {
+#[tokio::test]
+async fn test_cfi_at_4004() {
     // Should be the same as 4003
     let (mut f, mut stack, mut expected, expected_valid) = init_cfi_state();
 
@@ -447,11 +463,11 @@ fn test_cfi_at_4004() {
     f.raw.set_register("esi", 0xa97f229d);
     f.raw.set_register("edi", 0xb05cc997);
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_at_4005() {
+#[tokio::test]
+async fn test_cfi_at_4005() {
     let (mut f, mut stack, mut expected, expected_valid) = init_cfi_state();
 
     let frame1_rsp = Label::new();
@@ -470,11 +486,11 @@ fn test_cfi_at_4005() {
     f.raw.set_register("esi", 0x0fb7dc4e);
     f.raw.set_register("edi", 0x40005510);
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_at_4006() {
+#[tokio::test]
+async fn test_cfi_at_4006() {
     let (mut f, mut stack, mut expected, expected_valid) = init_cfi_state();
 
     let frame0_ebp = Label::new();
@@ -497,12 +513,12 @@ fn test_cfi_at_4006() {
     f.raw.set_register("esi", 0x743833c9);
     f.raw.set_register("edi", 0x40005510);
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
 // Totally basic STACK WIN frame data, no weird stuff.
-#[test]
-fn test_stack_win_frame_data_basic() {
+#[tokio::test]
+async fn test_stack_win_frame_data_basic() {
     let mut f = TestFixture::new();
     let symbols = [
         "STACK WIN 4 aa85 176 0 0 4 10 4 0 1",
@@ -544,7 +560,7 @@ fn test_stack_win_frame_data_basic() {
         .set_register("esp", stack.start().value().unwrap() as u32);
     f.raw.set_register("ebp", 0xf052c1de);
 
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 2);
 
     {
@@ -591,8 +607,8 @@ fn test_stack_win_frame_data_basic() {
 }
 
 // Totally basic STACK WIN frame data, no weird stuff.
-#[test]
-fn test_stack_win_frame_data_overlapping() {
+#[tokio::test]
+async fn test_stack_win_frame_data_overlapping() {
     // Same as frame_data_basic but there are extra entries which technically overlap
     // with this one, but in a way that is easily disambiguated by preferring the
     // one with the higher base address. This happens frequently in real symbol files.
@@ -647,7 +663,7 @@ fn test_stack_win_frame_data_overlapping() {
         .set_register("esp", stack.start().value().unwrap() as u32);
     f.raw.set_register("ebp", 0xf052c1de);
 
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 2);
 
     {
@@ -694,8 +710,8 @@ fn test_stack_win_frame_data_overlapping() {
 }
 
 // Testing that grand_callee_parameter_size is properly computed.
-#[test]
-fn test_stack_win_frame_data_parameter_size() {
+#[tokio::test]
+async fn test_stack_win_frame_data_parameter_size() {
     let mut f = TestFixture::new();
 
     let module1_symbols = ["FUNC 1000 100 c module1::wheedle\n"];
@@ -756,7 +772,7 @@ fn test_stack_win_frame_data_parameter_size() {
     f.raw
         .set_register("ebp", frame0_ebp.value().unwrap() as u32);
 
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 3);
 
     {
@@ -818,3 +834,139 @@ fn test_stack_win_frame_data_parameter_size() {
         }
     }
 }
+
+#[tokio::test]
+async fn test_frame_pointer_overflow() {
+    // Make sure we don't explode when trying frame pointer analysis on a value
+    // that will overflow.
+
+    type Pointer = u32;
+    let stack_max: Pointer = Pointer::MAX;
+    let stack_size: Pointer = 1000;
+    let bad_frame_ptr: Pointer = stack_max;
+
+    let mut f = TestFixture::new();
+    let mut stack = Section::new();
+    let stack_start: Pointer = stack_max - stack_size;
+    stack.start().set_const(stack_start as u64);
+
+    stack = stack
+        // frame 0
+        .append_repeated(0, stack_size as usize); // junk, not important to the test
+
+    f.raw.eip = 0x7a100000;
+    f.raw.ebp = bad_frame_ptr;
+    f.raw.esp = stack.start().value().unwrap() as Pointer;
+
+    let s = f.walk_stack(stack).await;
+    assert_eq!(s.frames.len(), 1);
+
+    // As long as we don't panic, we're good!
+}
+
+#[tokio::test]
+async fn test_frame_pointer_overflow_nonsense_32bit_stack() {
+    // same as test_frame_pointer_overflow, but we're going to abuse the fact
+    // that rust-minidump prefers representing things in 64-bit to create
+    // impossible stack addresses that overflow 32-bit integers but appear
+    // valid in 64-bit. By doing this memory reads will "succeed" but
+    // pointer math done in the native pointer width will overflow and
+    // everything will be sad.
+
+    type Pointer = u32;
+    let pointer_size: u64 = std::mem::size_of::<Pointer>() as u64;
+    let stack_max: u64 = Pointer::MAX as u64 + pointer_size * 2;
+    let stack_size: u64 = 1000;
+    let bad_frame_ptr: u64 = Pointer::MAX as u64 - pointer_size;
+
+    let mut f = TestFixture::new();
+    let mut stack = Section::new();
+    let stack_start: u64 = stack_max - stack_size;
+    stack.start().set_const(stack_start as u64);
+
+    stack = stack
+        // frame 0
+        .append_repeated(0, 1000); // junk, not important to the test
+
+    f.raw.eip = 0x7a100000;
+    f.raw.ebp = bad_frame_ptr as u32;
+    f.raw.esp = stack.start().value().unwrap() as Pointer;
+
+    let s = f.walk_stack(stack).await;
+    assert_eq!(s.frames.len(), 1);
+
+    // As long as we don't panic, we're good!
+}
+
+#[tokio::test]
+async fn test_frame_pointer_barely_no_overflow() {
+    // This is test_tradition but with the all the values pushed
+    // as close to the upper memory boundary as possible, to confirm that
+    // our code doesn't randomly overflow *AND* isn't overzealous in
+    // its overflow guards.
+
+    let mut f = TestFixture::new();
+    let mut stack = Section::new();
+
+    type Pointer = u32;
+    let pointer_size: Pointer = std::mem::size_of::<Pointer>() as Pointer;
+    let stack_max: Pointer = Pointer::MAX;
+    let stack_size: Pointer = pointer_size * 3;
+
+    let stack_start: Pointer = stack_max - stack_size;
+    let return_address: Pointer = 0x7b302000;
+    stack.start().set_const(stack_start as u64);
+
+    let frame0_fp = Label::new();
+    let frame1_sp = Label::new();
+    let frame1_fp = Label::new();
+
+    stack = stack
+        // frame 0
+        .mark(&frame0_fp)
+        .D32(&frame1_fp) // caller-pushed %rbp
+        .D32(return_address) // actual return address
+        // frame 1
+        .mark(&frame1_sp)
+        .mark(&frame1_fp) // end of stack
+        .D32(0);
+
+    f.raw.eip = 0x7a100000;
+    f.raw.ebp = frame0_fp.value().unwrap() as Pointer;
+    f.raw.esp = stack.start().value().unwrap() as Pointer;
+
+    let s = f.walk_stack(stack).await;
+    assert_eq!(s.frames.len(), 2);
+
+    {
+        // To avoid reusing locals by mistake
+        let f0 = &s.frames[0];
+        assert_eq!(f0.trust, FrameTrust::Context);
+        assert_eq!(f0.context.valid, MinidumpContextValidity::All);
+        if let MinidumpRawContext::X86(ctx) = &f0.context.raw {
+            assert_eq!(ctx.ebp, frame0_fp.value().unwrap() as Pointer);
+        } else {
+            unreachable!();
+        }
+    }
+
+    {
+        // To avoid reusing locals by mistake
+        let f1 = &s.frames[1];
+        assert_eq!(f1.trust, FrameTrust::FramePointer);
+        if let MinidumpContextValidity::Some(ref which) = f1.context.valid {
+            assert!(which.contains("eip"));
+            assert!(which.contains("esp"));
+            assert!(which.contains("ebp"));
+        } else {
+            unreachable!();
+        }
+        if let MinidumpRawContext::X86(ctx) = &f1.context.raw {
+            assert_eq!(ctx.eip, return_address);
+            assert_eq!(ctx.esp, frame1_sp.value().unwrap() as Pointer);
+            assert_eq!(ctx.ebp, frame1_fp.value().unwrap() as Pointer);
+        } else {
+            unreachable!();
+        }
+    }
+}
